import os
import time
import random
import logging
import asyncio
from types import CoroutineType
from typing import Any
from urllib.parse import urlparse
import requests
from warp_beacon.scraper.instagram.instagram import InstagramScraper
from pydub import AudioSegment
import speech_recognition as sr
from playwright.async_api import async_playwright, Page

class CaptchaSolver(object):
	TIMEOUT_STANDARD = 7
	TIMEOUT_SHORT = 1
	TIMEOUT_DETECTION = 0.05
	TEMP_DIR = "/tmp"

	scraper = None
	proxy_config = None

	def __init__(self, scraper: InstagramScraper) -> None:
		self.scraper = scraper
		if self.scraper.proxy:
			dsn = self.scraper.proxy.get("dsn", "")
			self.proxy_config = self.parse_proxy_from_dsn(dsn)

	def parse_proxy_from_dsn(self, dsn: str) -> dict:
		parsed = urlparse(dsn)
		
		proxy_config = {
			"server": f"{parsed.scheme}://{parsed.hostname}:{parsed.port}",
		}
		
		if parsed.username and parsed.password:
			proxy_config["username"] = parsed.username
			proxy_config["password"] = parsed.password

		return proxy_config

	async def _patch_page(self, page: CoroutineType[Any, Any, Page]):
		await page.add_init_script("""() => {
			Object.defineProperty(navigator, 'webdriver', { get: () => undefined });
			window.chrome = { runtime: {} };
			Object.defineProperty(navigator, 'languages', { get: () => ['en-US', 'en'] });
			Object.defineProperty(navigator, 'plugins', { get: () => [1,2,3,4,5] });
		}""")

	async def solve_audio_captcha(self, page: CoroutineType[Any, Any, Page]) -> None:
		logging.info("Processing audio captcha ..")
		mp3_path = os.path.join(self.TEMP_DIR, f"{random.randrange(1,1000)}.mp3")
		wav_path = os.path.join(self.TEMP_DIR, f"{random.randrange(1,1000)}.wav")
		try:
			await page.click('button[aria-label=\"Get an audio challenge\"]')
			time.sleep(0.3)
			await page.wait_for_selector('audio', timeout=10000)

			audio_src = await page.get_attribute('audio > source', 'src')
			audio_content = requests.get(audio_src, timeout=60).content
			with open(mp3_path, 'wb') as f:
				f.write(audio_content)

			if not os.path.exists(mp3_path):
				logging.error("MP3 file not downloaded!")
				return

			sound = AudioSegment.from_mp3(mp3_path)
			sound.export(wav_path, format='wav')

			recognizer = sr.Recognizer()
			with sr.AudioFile(wav_path) as source:
				audio = recognizer.record(source)

			try:
				text = recognizer.recognize_google(audio)
				logging.info("Detected text '%s'", text)
			except sr.UnknownValueError:
				logging.error("Failed to detect text!")
				text = ''

			if text:
				await page.fill('input[type=\"text\"]', text)
				await page.press('input[type=\"text\"]', 'Enter')
				logging.info("Audio captcha solved!")
		except Exception as e:
			logging.error("Exception in captcha audio solve!")
			logging.exception(e)
		finally:
			if os.path.exists(mp3_path):
				os.unlink(mp3_path)
			if os.path.exists(wav_path):
				os.unlink(wav_path)

	async def solve_challenge(self, challenge_url: str) -> None:
		async with async_playwright() as p:
			browser = None
			try:
				browser = await p.chromium.launch(
					headless=True,
					args=[
						"--no-sandbox",
						"--disable-blink-features=AutomationControlled",
						"--disable-infobars",
						"--disable-dev-shm-usage"
					],
					proxy=self.proxy_config
				)
				context = await browser.new_context(
					viewport={"width": 1280, "height": 800},
					user_agent="Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
					java_script_enabled=True,
					locale="en-US"
				)

				page = await context.new_page()
				await self._patch_page(page)

				await page.goto(challenge_url)

				# finding iframe with captcha
				frame_element = await page.wait_for_selector('iframe[src*=\"recaptcha\"]')
				frame = await frame_element.content_frame()
				time.sleep(0.1)

				# checkbox click
				checkbox = await frame.wait_for_selector('#recaptcha-anchor', timeout=10000)
				await checkbox.click()

				# waiting for frame with task
				await asyncio.sleep(3)

				# checking if iframe with task exists
				frames = page.frames
				challenge_frame = None
				for f in frames:
					if '/recaptcha/' in f.url and 'bframe' in f.url:
						challenge_frame = f
						break

				if not challenge_frame:
					logging.info("Captcha solved!")
				else:
					await self.solve_audio_captcha(challenge_frame)
			except Exception as e:
				logging.error("Exception in solver!")
				logging.exception(e)

			if browser:
				await asyncio.sleep(10)
				await browser.close()

	def run(self, challenge_url: str) -> None:
		asyncio.run(self.solve_challenge(challenge_url))


import os

from warp_beacon.telegram.bot import Bot

#import logging

def main() -> None:
	bot = Bot(
		tg_bot_name=os.environ.get("TG_BOT_NAME", default=None),
		tg_token=os.environ.get("TG_TOKEN", default=None),
		tg_api_id=os.environ.get("TG_API_ID", default=None),
		tg_api_hash=os.environ.get("TG_API_HASH", default=None)
	)
	bot.start()


import logging
from typing import Union
from bs4 import NavigableString, Tag, Comment
from bs4 import BeautifulSoup

CAPTION_LENGTH_LIMIT = 130

class CaptionShortner(object):
	@staticmethod
	def strip_html(text: str) -> str:
		try:
			soup = BeautifulSoup(text, "html.parser")
			return soup.get_text()
		except Exception as e:
			logging.warning("Failed to stript HTML tags!")
			logging.exception(e)

		return text

	@staticmethod
	def need_short(text: str) -> bool:
		wo_html = CaptionShortner.strip_html(text)
		if len(wo_html) > CAPTION_LENGTH_LIMIT:
			return True
		return False

	@staticmethod
	def smart_truncate_html(html: str, limit: int = CAPTION_LENGTH_LIMIT) -> str:
		result = ""
		try:
			soup = BeautifulSoup(html, "html.parser")
			length = 0

			def walk(node: Union[NavigableString, Tag, Comment]) -> None:
				nonlocal result, length
				if length >= limit:
					return

				if isinstance(node, str):
					words = node.split()
					for word in words:
						if length + len(word) + 1 > limit:
							return
						if result and not result.endswith(" "):
							result += " "
							length += 1
						result += word
						length += len(word)
				elif isinstance(node, Tag):
					if node.name == '[document]':
						for child in node.children:
							walk(child)
						return

					for child in node.children:
						walk(child)
						if length >= limit:
							break

					result += f"</{node.name}>"

			walk(soup)
		except Exception as e:
			logging.warning("Fail in smart_truncate_html!")
			logging.exception(e)
		return result


import io
from typing import Optional

import logging

import time
import json

import yt_dlp

from warp_beacon.jobs.types import JobType
from warp_beacon.scraper.youtube.abstract import YoutubeAbstract
from warp_beacon.scraper.exceptions import NotFound, FileTooBig, Unavailable

class YoutubeMusicScraper(YoutubeAbstract):
	YT_MAX_RETRIES_DEFAULT = 3
	YT_PAUSE_BEFORE_RETRY_DEFAULT = 3
	YT_TIMEOUT_DEFAULT = 2
	YT_TIMEOUT_INCREMENT_DEFAULT = 60

	def _download(self, url: str, session: bool = True, thumbnail: Optional[io.BytesIO] = None, timeout: int = 0) -> list:
		res = []
		try:
			yt = self.build_yt(url, session=session)
			
			stream = yt.streams.get_audio_only()
			
			if not stream:
				raise NotFound("No suitable audio stream found")
			
			logging.info("Announced audio file size: '%d'", stream.filesize)
			if stream.filesize > 2e+9:
				logging.warning("Downloading size reported by YouTube is over than 2 GB!")
				raise FileTooBig("YouTube file is larger than 2 GB")
			logging.info("Operation timeout is '%d'", timeout)
			local_file = stream.download(
				output_path=self.DOWNLOAD_DIR,
				max_retries=0,
				timeout=timeout,
				skip_existing=False,
				filename_prefix='yt_download_'
			)
			logging.debug("Temp filename: '%s'", local_file)
			res.append({
				"local_media_path": self.rename_local_file(local_file),
				"performer": yt.author,
				"thumb": thumbnail,
				"canonical_name": yt.title,
				"media_type": JobType.AUDIO
			})
		except Exception:
			raise Unavailable("Content unavailable")

		return res

	def build_yt_dlp(self, timeout: int = 60) -> yt_dlp.YoutubeDL:
		auth_data = {}
		with open(self.YT_SESSION_FILE % self.account_index, 'r', encoding="utf-8") as f:
			auth_data = json.loads(f.read())
		time_name = str(time.time()).replace('.', '_')
		ydl_opts = {
			'socket_timeout': timeout,
			'outtmpl': f'{self.DOWNLOAD_DIR}/yt_download_{time_name}.%(ext)s',
			'format': 'bestaudio[ext=m4a]/bestaudio/best',
			'noplaylist': True,
			'keepvideo': False,
			'tv_auth': auth_data
		}

		if self.proxy:
			proxy_dsn = self.proxy.get("dsn", "")
			logging.info("Using proxy DSN '%s'", proxy_dsn)
			if proxy_dsn:
				ydl_opts["proxy"] = proxy_dsn

		return yt_dlp.YoutubeDL(ydl_opts)

	def _download_yt_dlp(self, url: str, timeout: int = 60, thumbnail: Optional[io.BytesIO] = None) -> list:
		res = []
		with self.build_yt_dlp(timeout) as ydl:
			info = ydl.extract_info(url, download=True)
			local_file = ydl.prepare_filename(info)
			logging.debug("Temp filename: '%s'", local_file)
			res.append({
				"local_media_path": local_file,
				"performer": info.get("uploader", "Unknown"),
				"thumb": thumbnail,
				"canonical_name": info.get("title", ''),
				"media_type": JobType.AUDIO
			})

		return res
